// Interface.C
// This file is the interface between the main program
// and the scan converter. 
// In this file the command line arguments are parsed
// ScanScene allocates the draw buffer and 
// renders the scan-converted polygons. 



// system includes
#include <ctype.h>
#include <iostream.h>

// local includes
#include "ivwraplib/SaveWindow.h"
#include "SCtypes.h"
#include "Utils.h"

// Inits for user interface class
// The defaults are set here 
Binary interface::debug = 1;
int interface::blowup = 0;

char *interface::filename = NULL;
char *interface::envname = NULL;


extern void ivScan (Environment *env,
               void setPixel (int x, int y, SbColor color, Environment *env));


// Effect: argument parsing for command-line options
// Requires: argc, argv
// Modifies: interface::debug, interface::filename, interface::envname
SbBool getArgs (int argc, char **argv)
{
  interface::envname = NULL;
  interface::filename = NULL;

  for (int k = 1; k < argc; k++) {
    if (!strcmp(argv[k], "-e")) {            // -e envfile
      if (k == argc - 1) return FALSE;
      interface::envname = argv[++k];
    } else if (!strcmp (argv[k], "-help"))   // usage
      return FALSE;		
    else {
      // error if more than one
      if (interface::filename) return FALSE;
      // assume arg is iv filename
      interface::filename = argv[k];
    }
  }
  // args seem ok
  return TRUE;
}

// Effect:  prints out a usage comment
void Usage (char **argv)
{
  fprintf(stderr, "usage: %s\n", argv[0]);
  fprintf(stderr, "\t[-e environ.iv]\n");
  fprintf(stderr, "\t[file.iv (- for stdin)]\n");
  exit(1);
}

// Requires: env->avpSize, env->viewVolume, env->sceneGraph
// Effect: Sets up the pixelData draw buffer
//         scan converts the polygons
//         outputs the draw buffer
// Modifies: env and everything in it
//
void ScanScene (Environment *env)
{
  // allocate framebuffer
  assert (!env->pixelData);
  env->pixelData = new GLuint[env->avpSize[0] * env->avpSize[1]];
  assert (env->pixelData);
  
  if (interface::debug) fprintf (stderr,"calling ivScan...\n");
  if ( interface::blowup ) fprintf (stderr,"pixel blowup is %d.\n",
				  interface::blowup);
  
  // do the scan conversion
  // Renders the Scan Lines
  if ( !interface::debug && !interface::blowup )
    ivScan ( env, setPixelBuffer );
  else if ( interface::debug && !interface::blowup )
    ivScan ( env, setPixelImmed );
  else {
    switch ( interface::blowup ) {
      case 4 : ivScan ( env, setPixel4 );
	       break;
      case 8 : ivScan ( env, setPixel8 );
	       break;
      case 16 : ivScan ( env, setPixel16 );
	       break;
      case 32 : ivScan ( env, setPixel32 );
	       break;
      default: fprintf (stderr, "Error, illegal blowup factor\n");
	       break;
      }
    }
  
  if ( !interface::blowup ) {
    // display resulting framebuffer
    glRasterPos2i (0, 0);
    glDrawPixels (env->avpSize[X], env->avpSize[Y], 
		GL_RGBA, GL_UNSIGNED_BYTE, env->pixelData);

    // save to .rgb file, and display iff debug == 1
    saveWindow ((unsigned long *) env->pixelData, 
	      env->avpSize[X], env->avpSize[Y],
	      interface::debug == 1 );
    }

  
  // free the framebuffer 
  delete [] env->pixelData;
  env->pixelData = NULL;
}

// Effect:   The name says it all! Cleanup
//           Free  lights and reset counts numPoly, nextPoly
// Modifies: this
// Requires: curLights
void Environment::CleanUp ()
{
  nextPoly = numPoly = 0;
  // delete list of lights

  LightList *tmp = curLights;
  while (tmp) {
    curLights = curLights->next;
    delete tmp;
    tmp = curLights;
  }
  curLights = NULL;

}



