/* SCtypes.h
     Various important types used throughout this program.

     Main Classes:
       ViewVolume
       Environment
       interface
*/


#ifndef _SCANTYPES_
#define _SCANTYPES_
// They include the following types:.............

// system include files
#include <stdio.h>
// Inventor include files
#include <Inventor/SbBasic.h>
#include <Inventor/SbLinear.h>
#include <Inventor/SbBox.h>
#include <Inventor/SbColor.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoShape.h>
#include <Inventor/nodes/SoEnvironment.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoCoordinate4.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoLight.h>
#include <Inventor/nodes/SoSpotLight.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoDirectionalLight.h>

#include <assert.h>

#include "Types.h"

// tag for function parameters
#define OUTPUT	    // no-op

// typedef point, vector, color, radiance, 4x4 matrix
typedef SbBool  Binary;
typedef SbVec3f Point;
typedef Point  *PointP;
typedef SbVec2f Vector2D;
typedef SbVec3f Vector;
typedef Vector *VectorP;
typedef SbVec4f Plane;
typedef Plane  *PlaneP;
typedef SbColor Color;
typedef SbVec3f Radiance;
typedef SbBox3f Bbox;
typedef Bbox   *BboxP;
typedef void   *VoidP;
typedef SbMatrix Matrix;

// A point in 2d, continuous screen space
typedef SbVec2f Point2D;


#define X 0
#define Y 1
#define Z 2

#define R 0
#define G 1
#define B 2


// List of Lights
class LightList
{
public:
  SoLight *light;
  SbMatrix matrix;
  LightList *next;
};



// Viewing frustum, along with camera orientation specifier.
struct ViewVolume
{
  Point eye;		    // Eye position, in WS
  Vector viewDir;	    // normalized View direction, in WS
  Vector upDir;		    // normalized Up vector, in WS
  float aspectRatio;	    // width / height on near clipping plane
  float nearDistance;	    // Distance to near clipping plane.
  float farDistance;	    // Distance to far clipping plane.
  float fovy;		    // Vertical field of view, in radians.
  SbRotation orientation;   // Required by certain iv constructors.
};

// Forward decs of things used in Environment
class EdgeRecList;
class Edge;
class Polygon;

// The environment we are scanning into.  Maintains the list
// of polygons, lights, etc.
class Environment {
public:

  void CleanUp ();

  // Constructor 
  Environment () {
    pixelData    = NULL;
    avpSize = SbVec2s (0, 0);
    wvpSize = SbVec2s (0, 0);
    nextPoly     = 0; 
    numPoly      = 0;
    curLights    = NULL;
    Polygons     = NULL;
    EdgeRecTable = NULL;
    EdgeTable    = NULL;
    sceneGraph   = NULL;
  }

  GLuint        *pixelData;
  SbVec2s       avpSize; // actual opengl viewport
  SbVec2s       wvpSize; // working scan viewport
  int           nextPoly;
  int           numPoly;
  ViewVolume    viewVolume;
  LightList     *curLights;
  Polygon       *Polygons;
  EdgeRecList   *EdgeRecTable; // This is the table of edges starting at
                               // a given y-valuesvalue.
			       // e.g. EdgeRecTable[4] is a linked list
			       //   of all edges starting at y = 4.
  Edge          *EdgeTable;
  SoSeparator   *sceneGraph;
  
};

/*
  Maintain various debugging flags, the blowup flag, etc.
  A catchall structure of the interface parameters of the run.
*/
class interface
{
 public:
  static Binary debug;
  static int blowup;
  static Binary out;
  static char *filename;
  static char *envname;
};


#endif /* !_SCANTYPES_ */









