// ScanWrap.C
//
// (Very Ugly) Setup for scan conversion. 
// The code here does the projection and lighting and blah-blah-blah
// The code in Scan.C does the real funky-cool scan line stuff.

// system includes
#include <stdio.h>

// inventor includes
#include <Inventor/SbBasic.h>
#include <Inventor/SbColor.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/SoPrimitiveVertex.h>

// local includes
#include "SCtypes.h"
#include "Utils.h"
#include "Scan.h"
#include "Interface.h"
#include "EdgeRec.h"
#include "EdgeRecList.h"
#include "Edge.h"

// forward declarations
void InitScanET (Environment *env);


void Primitives (Environment *env);

void PrimitiveCountCB (void *data, SoCallbackAction *,
			      const SoPrimitiveVertex *,
			      const SoPrimitiveVertex *,
			      const SoPrimitiveVertex *);
void PrimitiveCB (void *data, SoCallbackAction *,
			 const SoPrimitiveVertex *,
			 const SoPrimitiveVertex *,
			 const SoPrimitiveVertex *);
static SoCallbackAction::Response addLightCB (void *data, 
					      SoCallbackAction *action, 
					      const SoNode *light);


// Effect:  Draw the blow-up pixel grid on the screen.
static void drawGrid (Environment *env, int showCenters )
{
  // draw coordinate grid
  int i, j;
  SbColor color = Color (0.2F, 0.2F, 0.2F);
  GLuint rgba = color.getPackedValue();

  // draw vertical 1-pixel wide lines on actual viewport
  for ( i = 0 ; i <= env->wvpSize[X]; i++ )
    for ( j = 0 ; j < env->avpSize[Y]; j++ ) {
      glRasterPos2i( i * interface::blowup, j );
      glDrawPixels(1, 1, GL_RGBA, GL_UNSIGNED_BYTE,  &rgba);
      }

  // draw horizontal 1-pixel wide lines on actual viewport
  for ( i = 0 ; i <= env->wvpSize[Y]; i++ ) 
    for ( j = 0 ; j < env->avpSize[X]; j++ ) {
      glRasterPos2i( j, i * interface::blowup );
      glDrawPixels(1, 1, GL_RGBA, GL_UNSIGNED_BYTE,  &rgba);
      }

  if ( showCenters ) {
    // draw pixel centers
    color = Color (0.5F, 0.5F, 0.5F);
    rgba = color.getPackedValue();
    for ( i = 0 ; i <= env->wvpSize[X]; i++ ) 
      for ( j = 0 ; j <= env->wvpSize[Y]; j++ ) {
        glRasterPos2i( interface::blowup/2 + i * interface::blowup, 
		       interface::blowup/2 + j * interface::blowup );
        glDrawPixels(1, 1, GL_RGBA, GL_UNSIGNED_BYTE,  &rgba);
        }
    }
  glFlush();
} // end drawpixelgrid


// Effect: Takes env->sceneGraph, and scan-converts the scene by invoking
//         setPixel with the appropriate color for each pixel in the output.
// Modifies: env->EdgeTable, env->EdgeRecTable, env->Polygons, env->pixelData 
// Requires: env->wvpSize must contain size of viewport
//           env->viewVolume contains camera frustum description
//           env->PixelData contains enough space for output
//           env->sceneGraph points to Inventor sceneGraph
void ivScan ( Environment *env,
	      void setPixel (int x, int y, SbColor color, Environment *env))
{
// Implementation:
//         Sets up the env->EdgeTable with all the edges in the scene
//         Invokes InitScanET which sets up the EdgeRecTable with EdgeRecs
//         for each scanline in y.
//         Invokes RenderScene which scanconverts the scene

  assert (env->wvpSize[Y] > 0);
  int yres = env->wvpSize[Y];

  env->EdgeRecTable = new EdgeRecList[yres];
  assert (env->EdgeRecTable);

  // Get all the polygons that have to be rendered
  Primitives (env);

  // show the blownup pixel grid
  if ( interface::blowup )
    drawGrid ( env, 1 );

  // Set up the EdgeRec Y Buckets
  InitScanET (env);

  // Renders the Scan Lines
  RenderScene (env, setPixel);

  // remove cruft on grid lines
  if ( interface::blowup )
    drawGrid ( env, 0 );

  // Free polygons
  delete [] env->Polygons;
  env->Polygons = NULL;

  // Free edgelist
  delete [] env->EdgeTable;
  env->EdgeTable = NULL;

  // Free event buckets
  for (int y = 0; y < yres; y++) env->EdgeRecTable[y].Clear ();
  delete [] env->EdgeRecTable;
  env->EdgeRecTable = NULL;
  
}


// Effects: Finds the number of polygons (triangles in this case)
//          in the scene (The polygons are used to approximate shapes in 
//          SceneViewer). 
//          Allocates space and initializes the Polygons array 
//          in the environment.
// Modifies: env->Polygons, env->numPoly, env->currPoly
// Requires: env->sceneGraph properly set up.
//           env->EdgeRecTable have space allocated in it 
void Primitives (Environment *env)
{
  env->numPoly = 0;
  env->nextPoly = 0;

  // Count the number of polygons
  SoCallbackAction *action = new SoCallbackAction;
  action->addTriangleCallback (SoShape::getClassTypeId(), 
			       &PrimitiveCountCB, env);
  action->addPreCallback (SoLight::getClassTypeId(), 
			  addLightCB, env);
  action->apply (env->sceneGraph);
  delete action;

  fprintf (stderr, "Counted %d primitives...\n", env->numPoly);

  // Allocate the array of polygons
  env->Polygons = new Polygon[env->numPoly];
  assert (env->Polygons);

  // Get Polygons and Lights in scene
  // PrimitiveCB sets up the polygons in env->Polygons
  action = new SoCallbackAction;
  action->addTriangleCallback (SoShape::getClassTypeId(),
			       &PrimitiveCB, env);
  action->apply (env->sceneGraph);
  delete action;


  if (interface::debug) 
    fprintf(stderr, "Returning the primitives...\n");
}



// Effect: Counts the number of polygons in the scene 
// Modifies: env->numPoly
// Requires: 
static void PrimitiveCountCB (void *data, SoCallbackAction *,
			    const SoPrimitiveVertex *,
			    const SoPrimitiveVertex *,
			    const SoPrimitiveVertex *)
{
  Environment *env = (Environment *) data;
  env->numPoly++;
}


// Effect:  Creates the local list of lights in the scene 
//          Adds lights from scene to env->curLights
// Modifies: env->curLights
// Requires: data is the pointer to env 
static SoCallbackAction::Response addLightCB (void *data, 
					      SoCallbackAction *action, 
					      const SoNode *light)
{
  LightList *lightNode = new LightList;
  assert  (lightNode);
  Environment *env = (Environment *) data;
  

  lightNode->next = env->curLights;
  lightNode->matrix = action->getModelMatrix();
  lightNode->light = (SoLight *) light;
  env->curLights = lightNode;
  return SoCallbackAction::CONTINUE;
}


// Modifies:   none
// Effect:  returns the color at point <pt> using a local lighting model
//          Light the point pt with all the lights in the scene.  
//          We use a local lighting model for this problem set;
//          i.e., the algorithm assumes that all lights are visible from this
//          point and does not handle occlusion.
//          In Problem Set 3 we will explore a global lighting model 
//          in the context of a ray tracer.  Also in this problem set we 
//          are only using the diffuse material properties to color 
//          vertices. A more complete shading model will be explored 
//          in Problem Set 3.
Color LightPoint (Environment *e, Vector pt, Vector norm, Color diffuse, Color,
		  Color, Color, float, float)
{

  Color accum = Color (0, 0, 0);
  norm.normalize ();

  for (LightList *light = e->curLights; light; light = light->next) {
    if (light->light->isOfType (SoDirectionalLight::getClassTypeId())) {
      Vector dirWS;
      light->matrix.multDirMatrix (((SoDirectionalLight *) light->light)->
		   direction.getValue(), dirWS);
      dirWS.normalize ();
      accum += (fmax (-norm.dot(dirWS), 0) *
		light->light->intensity.getValue() *
		CompMult (diffuse, light->light->color.getValue ()));
    } else if (light->light->isOfType (SoPointLight::getClassTypeId())) {
      Vector dirWS;
      light->matrix.multVecMatrix (((SoPointLight *) light->light)->
				   location.getValue(), dirWS);
      dirWS -= pt; // this is actually backwards-- no '-' below
      dirWS.normalize ();
      accum += (fmax (norm.dot (dirWS), 0) * 
		light->light->intensity.getValue () *
		CompMult (diffuse, light->light->color.getValue ()));
    }
  }

  // clip accum if greater than 1
  accum[0] = fmin (accum[0], 1);
  accum[1] = fmin (accum[1], 1);
  accum[2] = fmin (accum[2], 1);
  
  for (int c = 0; c < 3; c++) {
    if (accum[c] < 0.0) {
      printf ("accum < 0. %f %d\n", accum[c], c);
      accum[c] = 0.0;
    }
  }

  return accum;
}

// Effect: Viewport Transformation
//         sspt is set to the screen space coordinates of point espt
// Modifies: sspt
// Requires: e->viewVolume to be setup correctly
void FrustumToScreen (Environment *e, Point ppt, Point2D &sspt)
{
  float yscale = 2;
  if (e->viewVolume.aspectRatio < 1) yscale /= e->viewVolume.aspectRatio;

  float xscale = yscale * e->viewVolume.aspectRatio;

  // map point to screen space in actual viewport
  sspt[X] = (e->avpSize[X] * ppt[X] / xscale) + 0.5 * e->avpSize[X];
  sspt[Y] = (e->avpSize[Y] * ppt[Y] / yscale) + 0.5 * e->avpSize[Y];
}


// Requires: data is pointer to Env* passed when the callback was set up
//           env->Polygons has enough space allocated for all polygons
//           in scene. 
// Modifies:
//       env->Polygons[env->numPoly], env->numPoly
// Effect: 
//       Sets up a polygon in the env->Polygons array with the right
//       color, eye space coordinates, and screen space coordinates
//       This procedure is invoked with the 3 coordinates of 
//       the polygon v1, v2, v3. The material properties of each vertex
//       (from Inventor) are used to color the vertices of
//       the polygon. The current transformation is used to find the 
//       screen coordinates, and eye coordinates. 
static void PrimitiveCB (void *data, SoCallbackAction *action,
		       const SoPrimitiveVertex *v1, 
		       const SoPrimitiveVertex *v2, 
		       const SoPrimitiveVertex *v3)
{
 
  Environment *env = (Environment *) data;

  Vector v[3], norm[3];
  int    mat[3];

  v[0] = v1->getPoint (); norm[0] = v1->getNormal ();
  mat[0] = v1->getMaterialIndex ();
  v[1] = v2->getPoint (); norm[1] = v2->getNormal ();
  mat[1] = v2->getMaterialIndex ();
  v[2] = v3->getPoint (); norm[2] = v3->getNormal ();
  mat[2] = v3->getMaterialIndex ();
  

  SbMatrix ModelMatrix = action->getModelMatrix ();
  SbMatrix ProjMatrix  = action->getProjectionMatrix ();
  SbMatrix ViewMatrix  = action->getViewingMatrix ();
  SbMatrix TotalMatrix = ModelMatrix * ViewMatrix * ProjMatrix;

  SbMatrix EyeMatrix   = ModelMatrix * ViewMatrix;


  // for each of the three points of the polygon:
  // set its Eye Space coordinates, Screen Space coordinates,
  // and Color (using diffuse component of polygon material)
  Point    espt[POLYNUMVERTS], tmp[POLYNUMVERTS], wspt[POLYNUMVERTS];
  Point    wsnorm[POLYNUMVERTS], esnorm[POLYNUMVERTS];
  Point2D  sspt[POLYNUMVERTS];
  Color    col[POLYNUMVERTS];

  for (int i = 0; i < POLYNUMVERTS; i++) {

    // Get the eye space point, and normal for shading 
    ModelMatrix.multVecMatrix (v[i], wspt[i]);
    ModelMatrix.multDirMatrix (norm[i], wsnorm[i]);

    // Projected point
    TotalMatrix.multVecMatrix (v[i], tmp[i]);

    // Eye space point and normal
    EyeMatrix.multVecMatrix (v[i], espt[i]);
    EyeMatrix.multDirMatrix (norm[i], esnorm[i]);

    // Get Screen Space coordinates 
    // Warning: This is not a good thing to do. When you
    // get close to a polygon, if some vertex of the polygon lies
    // BEHIND your viewpoint, it will get incorrectly projected 
    // to the front of your viewpoint. This will result in erroneous
    // results. The correct solution requires first clipping edges
    // to the view frustum, and then projecting the resulting vertices.
    // This however, could result in more than 3 vertices per polygon.
    FrustumToScreen (env, tmp[i], sspt[i]);

    Color ambient, diffuse, specular, emission;
    float shininess, transparency;
    
    // Get the material properties
    action->getMaterial (ambient, diffuse, specular, emission, 
			 shininess, transparency, mat[i]);
    
    // Find the color of the vertex using a local shading model
    col[i] = LightPoint (env, wspt[i], wsnorm[i], diffuse, 
			 ambient, specular, emission,
			 shininess, transparency);
  }
  

  // if blownup pixels in effect, divide
  if ( interface::blowup ) {
    for (i = 0; i < POLYNUMVERTS; i++) {
      sspt[i][X] /= interface::blowup;
      sspt[i][Y] /= interface::blowup;
    }
  }

  // If any of the points lies outside the viewfrustum 
  // remove the polygon
  for (i = 0; i < POLYNUMVERTS; i++) {
    if ((sspt[i][X] < 0) || ((int)(sspt[i][X]+0.5) > (env->wvpSize[X]-0)) ||
	(sspt[i][Y] < 0) || ((int)(sspt[i][Y]+0.5) > (env->wvpSize[Y]-0)))
      return;
  }

  // Build the plane equation for this polygon and if it is back-face
  // culling eliminate the polygon 
  Vector pnorm = (espt[1] - espt[0]).cross (espt[2] - espt[0]);
  if (pnorm.length () < 1.e-10) {
    // points too close 
    return;
  }
  pnorm.normalize();
  Plane p = Plane (pnorm[0], pnorm[1], pnorm[2], -pnorm.dot(espt[0]));
  // Does eye pt lie in positive half-plane of plane p
  // pdist = p[0] * eyept[0] + p[1] * eyept[1] + p[2] * eyept[2] + 
  //         p[3] 
  // but eyept = Point (0, 0, 0);
  // Therefore, pdist = p[3];
  float pdist = p[3];
  if (pdist <= 0)  return;

  env->Polygons[env->nextPoly].Init (espt, sspt, col, esnorm);
  env->nextPoly++;
}



// Effect: Sets up env->EdgeRecTable with EdgeRecs that start at
//         each y scanline. The EdgeRecs are initialized with the
//         correct start and end SS and ES coordinates, and colors. 
//         They are also set up with the appropriate deltas for 
//         x, z, h, color etc. 
// Modifies: env->EdgeTable, env->Polygons
// Requires: env set up as earlier. env->Polygons should be allocated and set
static void InitScanET (Environment *env)
{
  PolyVert tmp;

  Polygon *P = env->Polygons;
  
  // Create array of edges, which is to hold all edges in the scene
  Edge *ET = env->EdgeTable = new Edge[env->numPoly * POLYNUMEDGES];
  assert (env->EdgeTable);

  for (int j = 0; j < env->numPoly; j++) {
    Point2D s[2];
    Point   w[2];
    Color   c[2];

    // sort the points of the polygon on <Y, X>
    P[j].SortPts ();

    // Make the edge
    PolyVert pvtemp0 = P[j].pt(0);
    PolyVert pvtemp1 = P[j].pt(1);
    PolyVert pvtemp2 = P[j].pt(2);

    ET[POLYNUMEDGES * j].Init     (pvtemp0, pvtemp1, &P[j]);
    ET[POLYNUMEDGES * j + 1].Init (pvtemp0, pvtemp2, &P[j]);
    ET[POLYNUMEDGES * j + 2].Init (pvtemp1, pvtemp2, &P[j]);

    for (int edge = 0; edge < POLYNUMEDGES; edge++) {
      // Create a new EdgeRec
      EdgeRec *se = new EdgeRec;
      assert (se);

      //don't cast these to integers
      float ystart = ET[POLYNUMEDGES * j + edge].ss[0][Y];
      float yend = ET[POLYNUMEDGES * j + edge].ss[1][Y];
      float xstart = ET[POLYNUMEDGES * j + edge].ss[0][X];
      float xend = ET[POLYNUMEDGES * j + edge].ss[1][X];

      if ((int)(ystart+0.5) != (int)(yend+0.5)) {
	// If the edge is not a horizontal edge
	// Initialize the EdgeRec
	se->Init (&ET[POLYNUMEDGES * j + edge]);

	// Put element in proper place on edge table
	env->EdgeRecTable[(int)(ystart+0.5)].Insert (se);

	if (interface::blowup) { // show the start, end vertices
	  drawLine((int)(xstart * (float)interface::blowup),
		   (int)(ystart * (float)interface::blowup),
		   (int)(xend * (float)interface::blowup),
		   (int)(yend * (float)interface::blowup), 
		   Color (1., 1., 1.), env);
	}
	else if (interface::debug) { 
	  // draw edge on screen.
	  drawLine(xstart, ystart, xend, yend, Color (1., 1., 1.), env);
	}
      }
    } // for edge = each edge of polygon j
  } // for j = each polygon 

  // Sort the elements in each bucket
  for (j = 0; j < env->wvpSize[Y]; j++)
      env->EdgeRecTable[j].Sort ();
}


// Effect: Sorts the vertices of the polygon such that their screen space
//         coordinates are sorted on <Y, X>
//         i.e. p[0].sspt<Y, X> <= p[1].sspt<Y, X> <= p[2].sspt<Y, X>
// Modifies: p[0..2]
// Requires: p[0..2] are setup with the right values 
void Polygon::SortPts () 
{
  // Rearrange the pts so that v[0] < v[1] < v[2],
  // where order is defined by <SScoords[Y], SScoords[X]>
  Point2D v[3], tmpv;
  PolyVert tmp;

  for (int i = 0; i < 3; i++) v[i]= p[i].sspt;

  if ((v[0][Y] > v[1][Y]) || 
      ((v[0][Y] == v[1][Y]) && (v[0][X] > v[1][X]))) {
    // exchange pt0 and pt1
    tmp = p[1];     tmpv = v[1];
    p[1] = p[0];    v[1] = v[0];
    p[0] = tmp;     v[0] = tmpv;
  }
  if ((v[0][Y] > v[2][Y]) ||
      ((v[0][Y] == v[2][Y]) && (v[0][X] > v[2][X]))) {
    // move pt2 to pt0 and shift pt0, and pt1 up
    tmp = p[2];   tmpv = v[2];
    p[2] = p[1];  v[2] = v[1]; 
    p[1] = p[0];  v[1] = v[0]; 
    p[0] = tmp;   v[0] = tmpv;
  }
    
  if ((v[1][Y] > v[2][Y]) || 
      ((v[1][Y] == v[2][Y]) && (v[1][X] > v[2][X]))) { 
    // exchange pt2 and pt1
    tmp = p[2];   tmpv = v[2];
    p[2] = p[1];  v[2] = v[1]; 
    p[1] = tmp;   v[1] = tmpv;
  }

  for (i = 0; i < 3; i++) v[i]= p[i].sspt;

  assert 
    ((v[0][Y] < v[1][Y]) || ((v[0][Y] == v[1][Y]) && (v[0][X] <= v[1][X])));
  assert 
    ((v[0][Y] < v[2][Y]) || ((v[0][Y] == v[2][Y]) && (v[0][X] <= v[2][X])));
  assert 
    ((v[1][Y] < v[2][Y]) || ((v[1][Y] == v[2][Y]) && (v[1][X] <= v[2][X])));
}





